/** @preserve
 * Xcel Energy Electric Outage Map.
 * © 2010 Colin Snover. Released under MIT license.
 */
(function (window, undefined) {
	/**
	 * Name of exposed class in the global namespace.
	 * @type string
	 */
	var ns = 'OutageMap';

	/**
	 * @typedef {{name: string, latitude: number, longitude: number, zoom: number, outlines: Array.<string>}}
	 */
	var AreaConfigurationObject;

	/**
	 * @typedef {{count: string, created: string, lang: string, results: YQLQueryResult}}
	 */
	var YQLQuery;

	/**
	 * @typedef {{result: Object.<string, OutageInformationObject>}}
	 */
	var YQLQueryResult;

	/**
	 * Optional additional options to pass to the map.
	 * titleElement: A container element used to display the current map area’s
	 * name.
	 * summaryElement: A container element used to display the current map’s
	 * outage summary.
	 * modificationDateElement: A container element used to display the current
	 * map’s last modification date.
	 * @typedef {{titleElement: Element, summaryElement: Element, modificationDateElement: Element}}
	 */
	var OutageMapOptions;

	/**
	 * The conversion through YQL causes all data types to become strings, and
	 * causes arrays with single values to become scalar. Gross, but not much
	 * can be done about it.
	 * @typedef {{reportDate: (string|Array.<string>), numReports: string,
	 * numAffected: string, hasCrewAssigned: string, hasCrewOnsite: string,
	 * eta: string, latitude: string, longitude: string, communities: string }}
	 */
	var OutageInformationObject;

	(function () {
		/**
		 * @type {Function}
		 */
		var origParse = Date.parse;

		/**
		 * Parses an ISO-8601 datetime timestamp and returns it in milliseconds
		 * since Unix epoch. ES5 supports this natively.
		 * @param {string} date
		 * @return {number|NaN}
		 */
		Date.parse = function (date) {
			var timestamp = origParse(date), minutesOffset = 0, struct;
			if (isNaN(timestamp) && (struct = /(\d{4})-?(\d{2})-?(\d{2})(?:[T ](\d{2}):?(\d{2}):?(\d{2})(?:\.(\d{3,}))?(?:(Z)|([+\-])(\d{2})(?::?(\d{2}))?))/.exec(date))) {
				if (struct[8] !== 'Z') {
					minutesOffset = +struct[10] * 60 + (+struct[11]);

					if (struct[9] === '+') {
						minutesOffset = 0 - minutesOffset;
					}
				}

				timestamp = Date.UTC(+struct[1], +struct[2] - 1, +struct[3], +struct[4], +struct[5] + minutesOffset, +struct[6], +struct[7].substr(0, 3));
			}

			return timestamp;
		};
	}());

	/**
	 * Generate the correct path for a fragment relative to the current
	 * document/base URI.
	 * @param {string} fragment
	 * @return {string}
	 */
	function path(fragment) {
		var div = document.createElement('div'), href;

		// IE will not give us an absolute URL without innerHTML…
		div.innerHTML = '<a href="' + fragment + '"></a>';
		href = div.getElementsByTagName('a')[0].href;

		// …but it will leak memory if we aren’t careful.
		div.innerHTML = '';
		div = null;

		return href;
	}

	/**
	 * Proxy function.
	 * @param {Object} obj Object to bind as ‘this’.
	 * @param {Function} fn Function to call.
	 * @param {...*} var_args Arguments to pass to the end of the proxied
	 * function.
	 * @return {Function}
	 */
	function proxy(obj, fn, var_args) {
		var slice = Array.prototype.slice,
			args = slice.call(arguments, 2);
		return function () {
			return fn.apply(obj || this, Array.prototype.concat.call(slice.call(arguments), args));
		};
	}

	/**
	 * Performs a JSONP request.
	 * @param {string} url URL.
	 * @param {Object} data Query string parameters.
	 * @param {Function} callback Callback.
	 */
	var jsonp = (function () {
		/**
		 * @type {number}
		 */
		var counter = 0,
			body = document.body;

		return function (url, data, callback) {
			var puid = ns + (++counter),
				script = document.createElement('script');

			/**
			 * Sends data to the callback and cleans up after a JSONP request.
			 * @param {Object} data Data from the JSONP request.
			 */
			function handler(data) {
				callback(data);

				window[puid] = undefined;

				try {
					delete window[puid];
				}
				catch (e) {}

				body.removeChild(script);
			}

			/**
			 * Builds a query string from a hash map.
			 * @param {Object} data
			 * @return {string}
			 */
			function buildQueryString(data) {
				var query = '?',
					encode = encodeURIComponent;

				for (var i in data) {
					if (data.hasOwnProperty(i)) {
						query += encode(i) + '=' + encode(data[i]) + '&';
					}
				}

				query += 'callback=' + encode(puid);
				return query;
			}

			window[puid] = handler;

			script.src = url + buildQueryString(data);
			body.appendChild(script);
		};
	}());

	/**
	 * A new outage map.
	 * @param {string} Element ID to replace with a map.
	 * @param {string} Bing Maps credentials.
	 * @param {OutageMapOptions=} options Additional (optional) options.
	 */
	var OutageMap = this[ns] = function (id, credentials, options) {
		this.id = id;
		this.credentials = credentials;
		this.options = options || {};

		var map = this.map = new VEMap(this.id);
		map.SetCredentials(this.credentials);
		map.SetDashboardSize(VEDashboardSize.Small);
		map.AttachEvent('onendzoom', proxy(this, this._setClustering));

		// build default map options once on-demand
		if (!OutageMap.prototype._mapOptions) {
			var mapOptions = OutageMap.prototype._mapOptions = new VEMapOptions();
			mapOptions.EnableBirdseye = false;
		}
	};

	OutageMap.prototype = {
		/**
		 * @type {OutageMap}
		 */
		constructor: OutageMap,

		/**
		 * Element ID for this map.
		 * @type {string}
		 */
		id: undefined,

		/**
		 * Bing Maps credentials.
		 * @type {string}
		 */
		credentials: undefined,

		/**
		 * Additional options.
		 * @type {OutageMapOptions}
		 */
		options: undefined,

		/**
		 * An enumeration of NSPW, NSPM, PSCo, or SPS.
		 * @type {string}
		 */
		area: undefined,

		/**
		 * Map object.
		 * @type {VEMap}
		 */
		map: undefined,

		/**
		 * A shape layer upon which pushpins of each outage is placed.
		 * @type {VEShapeLayer}
		 */
		pinLayer: undefined,

		/**
		 * Standard options for all maps.
		 * @type {VEMapOptions}
		 */
		_mapOptions: undefined,

		/**
		 * Sets the area for the map to display.
		 * @param {OutageMap.AreaConfiguration} area
		 * @return {OutageMap}
		 */
		setArea: function (area) {
			/**
			 * Hides the default icon for a shape after it has been added.
			 * @param {VEShapeLayer} layer
			 */
			function onAreaShapeLoaded(layer) {
				layer.GetShapeByIndex(0).HideIcon();
			}

			if (this.area !== area) {
				var map = this.map,
					config = OutageMap.AreaConfiguration[area];

				if (!config) {
					throw new Error('Unknown area "' + area + '"');
				}

				// XXX there is some weird issue where if we set onLoadMap
				// function, the pinLayer stops working.
				map.LoadMap(
					new VELatLong(config.latitude, config.longitude),
					config.zoom,
					this._mapOptions
				);
				map.ClearInfoBoxStyles();

				for (var i = 0, j = config.outlines.length; i < j; ++i) {
					map.ImportShapeLayerData(
						new VEShapeSourceSpecification(
							VEDataType.GeoRSS,
							config.outlines[i],
							new VEShapeLayer()
						),
						onAreaShapeLoaded,
						false);
				}

				if (this.options.titleElement) {
					this.options.titleElement.innerHTML = config.name;
				}

				this.area = area;

				jsonp('http://query.yahooapis.com/v1/public/yql', {
					q: 'USE "' + OutageMap.TABLE_DEFINITION_URL + '" AS xcel.outagefeed; SELECT * FROM xcel.outagefeed WHERE area = "' + this.area + '";',
					format: 'json',
					_maxage: 600
				}, proxy(this, this.loadOutageData));
			}

			return this;
		},

		/**
		 * Loads outage information into the map & summary.
		 * @param {{query: YQLQuery}} data
		 */
		loadOutageData: function (data) {
			var descriptionTemplate = '<table><tr><th>Customers out</th><td>{out}</td></tr><tr><th>Report times</th><td>{start}</td></tr><tr><th>Restoration est.</th><td>{end}</td></tr><tr><th>Communities</th><td>{communities}</td></tr></table>',
				outages, outage, shape, pinColour, description,
				numOutages = 0, numAffected = 0,
				options = this.options;

			if (this.pinLayer) {
				this.map.DeleteShapeLayer(this.pinLayer);
			}

			this.pinLayer = new VEShapeLayer();

			if (data.query.results) {
				outages = data.query.results.result;
				for (var i in outages) {
					if (outages.hasOwnProperty(i)) {
						outage = outages[i];

						shape = new VEShape(VEShapeType.Pushpin, [ new VELatLong(+outage.latitude, +outage.longitude) ]);

						if (outage.numAffected <= 50) {
							pinColour = 'green';
						}
						else if (outage.numAffected < 500) {
							pinColour = 'yellow';
						}
						else if (outage.numAffected < 2500) {
							pinColour = 'orange';
						}
						else {
							pinColour = 'red';
						}

						description = descriptionTemplate
							.replace('{out}', +outage.numAffected)
							.replace('{start}', outage.reportDate instanceof Array ? outage.reportDate.join('<br>') : outage.reportDate)
							.replace('{end}', outage.eta)
							.replace('{communities}', outage.communities);

						if (outage.hasCrewOnsite === 'true') {
							description = '<p>Restoration in progress</p>' + description;
						}

						shape.SetCustomIcon('http://xcelenergy.com/OutageMap/images/32x32-' + pinColour + '.gif');
						shape.SetTitle('<h4>Outage Information</h4>');
						shape.SetDescription(description);

						numOutages += +outage.numReports;
						numAffected += +outage.numAffected;

						this.pinLayer.AddShape(shape);
					}
				}
			}

			this.map.AddShapeLayer(this.pinLayer);
			this._setClustering();

			if (options.summaryElement) {
				options.summaryElement.innerHTML = 'Total outages: {outages}<br>Num affected: {affected}'
					.replace('{outages}', numOutages)
					.replace('{affected}', numAffected);
			}

			if (options.modificationDateElement) {
				options.modificationDateElement.innerHTML = 'Last updated: {updated} (updates every 10 minutes)'
					.replace('{updated}', new Date(Date.parse(data.query.created)));
			}
		},

		/**
		 * Sets the title and description of cluster tooltips.
		 * @param {Array.<VEClusterSpecification>} clusters
		 */
		_drawClusters: function (clusters) {
			for (var i = 0, j = clusters.length; i < j; ++i) {
				var cluster = clusters[i], clusterShape = cluster.GetClusterShape();

				clusterShape.SetTitle('This cluster has {outages} outages in it.'.replace('{outages}', cluster.Shapes.length));
				clusterShape.SetDescription('Double click to zoom in for more details');
			}
		},

		/**
		 * Disables clustering for zoom levels 11+.
		 * @param {MapEvent=} event
		 */
		_setClustering: function (event) {
			if (!this.pinLayer) {
				return;
			}

			if (this.map.GetZoomLevel() < 11) {
				var options = new VEClusteringOptions();
				options.Callback = this._drawClusters;
				options.Icon = new VECustomIconSpecification();
				options.Icon.Image = 'http://xcelenergy.com/OutageMap/images/clustered.gif';
				this.pinLayer.SetClusteringConfiguration(VEClusteringType.Grid, options);
			}
			else {
				this.pinLayer.SetClusteringConfiguration(VEClusteringType.None);
			}
		}
	};

	/**
	 * Location of the YQL Open Data Tables schema for the outage feed.
	 * @type {string}
	 */
	OutageMap.TABLE_DEFINITION_URL = path('xcel.outagefeed.xml');

	/**
	 * Known map areas.
	 * @enum {AreaConfigurationObject}
	 */
	OutageMap.AreaConfiguration = {
		NSPW: {
			name: 'Michigan and Wisconsin',
			latitude: 45.1979,
			longitude: -90.7831,
			zoom: 7,
			outlines: [ 'WIOutline.xml' ]
		},

		NSPM: {
			name: 'Minnesota, North and South Dakota',
			latitude: 46.1979,
			longitude: -95.7831,
			zoom: 6,
			outlines: [ 'ND1Outline.xml', 'ND2Outline.xml', 'MNOutline.xml' ]
		},

		PSCo: {
			name: 'Colorado',
			latitude: 39.0279,
			longitude: -105.7831,
			zoom: 7,
			outlines: [ 'COOutline.xml' ]
		},

		SPS: {
			name: 'New Mexico and Texas',
			latitude: 35.1979,
			longitude: -102.7831,
			zoom: 6,
			outlines: [ 'TXOutline.xml' ]
		}
	};
}(window));